@@ -2,13 +2,16 @@
 
 """Stb-tester Python API: Automated GUI Testing for Set-Top Boxes.
 
-Copyright © 2013-2022 Stb-tester.com Ltd. All rights reserved.
+Copyright © 2013-2024 Stb-tester.com Ltd. All rights reserved.
 """
 
 from __future__ import annotations
 
+import enum
+import typing
 from contextlib import contextmanager
-from typing import Dict, Generator
+from dataclasses import dataclass
+from typing import Any, Callable, Generator, Optional, TypeVar
 
 import stbt_core  # open source APIs
 
@@ -24,12 +27,12 @@ from stbt_core import (
     crop,
     debug,
     detect_motion,
+    Differ,
     Direction,
     draw_text,
     find_file,
     for_object_repository,
     Frame,
-    FrameDiffer,
     FrameObject,
     frames,
     get_config,
@@ -39,6 +42,7 @@ from stbt_core import (
     Image,
     is_screen_black,
     Keyboard,
+    Keypress,
     last_keypress,
     load_image,
     load_mask,
@@ -56,8 +60,10 @@ from stbt_core import (
     MultiPress,
     NoVideo,
     ocr,
+    ocr_eq,
     OcrEngine,
     OcrMode,
+    PDU,
     Position,
     PreconditionError,
     press,
@@ -69,6 +75,7 @@ from stbt_core import (
     set_global_ocr_corrections,
     Size,
     TextMatchResult,
+    Transition,
     TransitionStatus,
     UITestError,
     UITestFailure,
@@ -82,6 +89,7 @@ from stbt_core import (
 # IDE linting & autocompletion.
 __all__ = stbt_core.__all__ + [
     "audio_chunks",
+    "AppleTV",
     "AudioChunk",
     "BadSyncPattern",
     "color_diff",
@@ -90,16 +98,24 @@ __all__ = stbt_core.__all__ + [
     "find_selection_from_background",
     "FIRST_FRAME_TIME",
     "get_rms_volume",
+    "Learning",
     "measure_av_sync",
+    "navigate_1d",
+    "navigate_grid",
+    "NavigationFailed",
+    "NavigationState",
+    "pdu",
     "play_audio_file",
     "prometheus",
     "RmsVolumeResult",
     "Roku",
+    "save_video_high_quality",
     "segment",
     "stop_job",
     "TEST_PACK_ROOT",
     "VolumeChangeDirection",
     "VolumeChangeTimeout",
+    "wait_for_content_start",
     "wait_for_volume_change",
 ]
 
@@ -112,6 +128,13 @@ from _stbt.imgutils import _frame_repr
 
 from . import prometheus
 
+if typing.TYPE_CHECKING:
+    import _stbt.imgutils
+
+    class TakesFrame(typing.Protocol):
+        def __call__(self, *, frame: Frame) -> Any:
+            ...
+
 
 FIRST_FRAME_TIME = None
 TEST_PACK_ROOT = None
@@ -129,6 +152,57 @@ def _raise_premium(api_name):
         "contact support@stb-tester.com for help" % api_name)
 
 
+class AppleTV:
+    """Control an AppleTV device using pyatv.
+
+    `pyatv`_ is an open source tool for controlling Apple TV devices. It uses
+    AirPlay and other network protocols supported by the Apple TV. This class
+    makes it easy to use pyatv in Stb-tester test scripts.
+
+    pyatv has additional dependencies that are not installed by default in the
+    Stb-tester v34 environment. For instructions on how to install these
+    dependencies, and how to configure the Apple TV's IP address, see
+    :doc:`pyatv`.
+
+    .. warning::
+
+        We have done the "integration" work to make pyatv work on the Stb-tester
+        Node, but we are not the authors of pyatv, and we don't provide support
+        for bugs in pyatv itself.
+
+    :param str address:
+        The IP address of the Apple TV device. If not specified, this is read
+        from "device_under_test.ip_address" in your :ref:`node-specific-config`.
+
+    .. _pyatv: https://pyatv.dev/
+    """
+
+    def __init__(self, address: Optional[str] = None):
+        _raise_premium("AppleTV")
+
+    def launch_app(self, name: str):
+        """Launches the specified app.
+
+        :param name: The name or bundle ID of the app (for example "Netflix" or
+            "com.netflix.Netflix").
+        """
+        return
+
+    def list_apps(self) -> dict[str, str]:
+        """Returns a dict of ``id: name`` with all the apps installed on the
+        Apple TV device.
+        """
+        return {}
+
+    def set_text(self, text: str):
+        """Set the text in a focused text field.
+
+        This only works if a text field is currently focused (for example a
+        search or login field).
+        """
+        return
+
+
 class AudioChunk(numpy.ndarray):
     """A sequence of audio samples.
 
@@ -484,11 +558,11 @@ class BadSyncPattern(RuntimeError):
 
 class FindSelectionFromBackgroundResult:
     def __init__(self, matched, region, mask_region, image, frame):
-        self.matched = matched
-        self.region = region
-        self.mask_region = mask_region
-        self.image = image
-        self.frame = frame
+        self.matched: bool = matched
+        self.region: "Region | None" = region
+        self.mask_region: "Region" = mask_region
+        self.image: "Image" = image
+        self.frame: "_stbt.imgutils.FrameT" = frame
 
     def __bool__(self):
         return self.matched
@@ -508,8 +582,11 @@ class FindSelectionFromBackgroundResult:
 
 
 def find_selection_from_background(
-        image, max_size, min_size=None, frame=None, mask=Region.ALL,
-        threshold=25, erode=True):
+        image: _stbt.imgutils.ImageT, max_size: tuple[int, int],
+        min_size: "tuple[int, int] | None" = None,
+        frame: _stbt.imgutils.FrameT | None = None,
+        mask: MaskTypes = Region.ALL, threshold: float = 25,
+        erode: bool = True) -> FindSelectionFromBackgroundResult:
 
     """Checks whether ``frame`` matches ``image``, calculating the region
     where there are any differences. The region where ``frame`` doesn't match
@@ -524,7 +601,7 @@ def find_selection_from_background(
 
     Unlike `stbt.match`, ``image`` must be the same size as ``frame``.
 
-    :param str|stbt.Image image:
+    :param image:
       The background to match against. It can be the filename of a PNG file on
       disk, or an image previously loaded with `stbt.load_image`.
 
@@ -532,35 +609,34 @@ def find_selection_from_background(
       is, the alpha channel is ANDed with ``mask``). This image must be the
       same size as ``frame``.
 
-    :param stbt.Size max_size:
+    :param max_size:
       The maximum size ``(width, height)`` of the differing region. If the
       differences between ``image`` and ``frame`` are larger than this in
       either dimension, the function will return a falsey result.
 
-    :param stbt.Size min_size:
+    :param min_size:
       The minimum size ``(width, height)`` of the differing region (optional).
       If the differences between ``image`` and ``frame`` are smaller than this
       in either dimension, the function will return a falsey result.
 
-    :param stbt.Frame frame:
+    :param frame:
       If this is specified it is used as the video frame to search in;
       otherwise a new frame is grabbed from the device-under-test. This is an
       image in OpenCV format (for example as returned by `stbt.frames` and
       `stbt.get_frame`).
 
-    :param str|numpy.ndarray|Mask|Region mask:
+    :param mask:
       A `Region` or a mask that specifies which parts of the image to
       analyse. This accepts anything that can be converted to a Mask using
       `stbt.load_mask`. See :doc:`masks`.
 
-    :type threshold: int
     :param threshold:
       Threshold for differences between ``image`` and ``frame`` for it to be
       considered a difference. This is a colour distance between pixels in
       ``image`` and ``frame``. 0 means the colours have to match exactly. 255
       would mean that even white (255, 255, 255) would match black (0, 0, 0).
 
-    :param bool erode:
+    :param erode:
       By default we pass the thresholded differences through an erosion
       algorithm to remove noise or small anti-aliasing differences. If your
       selection is a single line less than 3 pixels wide, set this to False.
@@ -580,7 +656,7 @@ def find_selection_from_background(
         ``find_selection_from_background``.
       * **frame** (`stbt.Frame`) – The video-frame that was analysed.
 
-    ``find_selection_from_background`` was added in v32.
+    ``stbt.find_selection_from_background`` was added in v32.
 
     Changed in v33: ``mask`` accepts anything that can be converted to a Mask
     using `load_mask` (previously it only accepted a `Region`).
@@ -642,9 +718,7 @@ def detect_pages(frame=None, candidates=None, test_pack_root=""):
     return iter([FrameObject()])
 
 
-def stop_job(reason=None):
-    # type: (str) -> None
-
+def stop_job(reason: Optional[str] = None) -> None:
     """Stop this job after the current testcase exits.
 
     If you are running a job with multiple testcases, or a
@@ -702,9 +776,11 @@ def color_diff(frame=None, *, background_color=None, foreground_color=None,
 
 def find_regions_by_color(
         color, *, frame=None, threshold=0.05, erode=False, mask=Region.ALL,
-        min_size=(20, 20), max_size=None):
+        min_size=(20, 20), max_size=None) -> list[Region]:
     """Find contiguous regions of a particular color.
 
+    For a guide to using this API see :doc:`find-regions-by-color`.
+
     :param stbt.Frame frame: The video frame to process.
     :param Color color: See the ``foreground_color`` parameter of `color_diff`.
     :param float threshold: See `color_diff`.
@@ -728,7 +804,8 @@ def find_regions_by_color(
 
 
 def segment(frame, *, region=Region.ALL, initial_direction=Direction.VERTICAL,
-            steps=1, narrow=True, light_background=False):
+            steps=1, narrow=True, light_background=False, merge_px=(0, 0)) \
+            -> list[Region]:
     """Segment (partition) the image into a list of contiguous foreground
     regions.
 
@@ -737,6 +814,8 @@ def segment(frame, *, region=Region.ALL, initial_direction=Direction.VERTICAL,
     binarization yourself (for example with `stbt.color_diff`) and pass the
     binarized image to ``segment``.
 
+    For a guide to using this API see :doc:`segment`.
+
     :param Frame frame:
       The video-frame or image to process.
 
@@ -772,10 +851,21 @@ def segment(frame, *, region=Region.ALL, initial_direction=Direction.VERTICAL,
       ``light_background=True`` if foreground pixels are dark (for example
       black text on a light background).
 
+    :param int|Size merge_px:
+      Merge nearby regions that are separated by a gap of this many pixels or
+      fewer. This is a tuple of (width, height). Regions are merged width-wise
+      during horizontal steps, and height-wise during vertical steps. For
+      example, to merge letters within the same word in horizontal text, use
+      ``merge_px=(5, 0)``.
+
+      Specifying a single integer means the same gap will be used for horizontal
+      and vertical steps. The default is ``(0, 0)`` which means no merging.
+
     :rtype: list[stbt.Region]
     :returns: A list of `stbt.Region` instances.
 
-    Added in v33.
+    | ``stbt.segment`` was added in v33.
+    | The ``merge_px`` parameter was added in v34.
     """
     _raise_premium("segment")
     return []
@@ -787,7 +877,7 @@ class Roku:
     This uses Roku's `External Control Protocol`_.
 
     To find the Roku's IP address and to enable the Roku's network control
-    protocol see <https://stb-tester.com/kb/roku>.
+    protocol see `Device Configuration: Roku <roku>`_.
 
     :param str address: IP address of the Roku.
 
@@ -843,7 +933,7 @@ class Roku:
         _raise_premium("Roku")
         yield
 
-    def query_apps(self) -> Dict[str, str]:
+    def query_apps(self) -> dict[str, str]:
         """Returns a dict of ``application_id: name`` with all the apps
         installed on the Roku device.
         """
@@ -857,3 +947,343 @@ class Roku:
         on the Roku.
         """
         _raise_premium("Roku")
+
+
+def save_video_high_quality(filename: str = "video-high-quality.mp4",
+                            duration_secs: int = 30):
+    """Record a "perceptually lossless" video from the device-under-test's HDMI
+    output.
+
+    This video is much larger in size than the "video.mp4" of the entire
+    test-run that is recorded automatically, so the duration is limited to a
+    maximum of 30 seconds.
+    """
+    _raise_premium("save_video_high_quality")
+
+
+def pdu(uri: str|None = None) -> PDU:
+    """Return a `PDU` (Power Distribution Unit) object to control the power to
+    the device-under-test.
+
+    :param str|None name:
+      The name of the PDU. This must match a PDU configured in the test-pack's
+      configuration files. If ``None`` then the name is taken from the
+      `device_under_test.power_outlet` configuration variable.
+
+    For more details see :doc:`pdu`.
+
+    ``stbt.pdu`` was added in v34.
+    """
+    _raise_premium("pdu")
+
+
+def wait_for_content_start(
+        keypress: Keypress | None = None,
+        spinner_fn: "TakesFrame | None" = None,
+        mask: MaskTypes = Region.ALL,
+        black_threshold: int | None = None,
+        motion_threshold: int | None = None,
+        consecutive_frames: int | str | None = None,
+        timeout_secs: float = 30,
+        threshold_db: float = 10.,
+    ) -> list[tuple[float, str, Any]]:
+    """Waits for content to start, taking performance measurements.
+
+    This looks for the following events in parallel::
+
+        keypress ─┬──> Screen goes black ──> No longer black ───┐
+                  ├──> Spinner appears ───> Spinner disappears ─┤
+                  ├───────────────> Sound ──────────────────────┤
+                  └───────────────> Motion ─────────────────────┴─> Done
+
+    It spawns a separate thread to detect each of these events in parallel.
+    It uses, respectively, `stbt.is_screen_black`, a custom spinner-detection
+    function provided by the caller, `stbt.wait_for_volume_change`, and
+    `stbt.wait_for_motion`.
+
+    :param keypress: The return value from `stbt.press` that
+        started content playback.
+
+    :param spinner_fn: A function that takes a ``frame`` argument and return
+        true if the spinner is visible in the frame, false otherwise.
+        Defaults to `None` (no spinner detection).
+
+    :param mask:
+        A mask used for motion and black-screen detection. This should mask out
+        the spinner and any UI elements superimposed on the video.
+
+    :param black_threshold:
+        See the ``threshold`` parameter of `stbt.is_screen_black`.
+
+    :param motion_threshold:
+        See the ``noise_threshold`` parameter of `stbt.wait_for_motion`.
+
+    :param consecutive_frames:
+        See the ``consecutive_frames`` parameter of `stbt.wait_for_motion`.
+
+    :param threshold_db:
+        See the ``threshold_db`` parameter of `stbt.wait_for_volume_change`.
+
+    :param timeout_secs: Number of seconds to wait for content to start
+        playing.
+
+    :return: A list of 3-element tuples (time, event_type, event) in time
+      order. Possible event types are: "press", "black", "not_black",
+      "spinner", "not_spinner", "audio", and "motion".
+
+    Example usage::
+
+        # after navigating to the right place in the UI so that the "play"
+        # button is focused and ready to be pressed...
+        keypress = stbt.press("KEY_OK")
+        events = wait_for_content_start(keypress, mask="mask-out-spinner.png")
+        start_time = keypress.start_time
+        for t, event, _ in events:
+            print("%.3f %s" % (t - start_time, event))
+        content_started = min(t - start_time for t, event, _ in events
+                              if event in ["not_black", "motion", "audio"])
+        print("Content started after %.3f seconds" % content_started)
+
+    This example code would print something like this::
+
+        0.000 press
+        0.406 black
+        2.606 motion
+        2.606 not_black
+        2.925 audio
+        Content started after 2.606 seconds
+
+    Added in v34.
+    """
+    _raise_premium("wait_for_content_start")
+    return []
+
+
+class Learning(enum.Enum):
+
+    #: Don't learn the menu structure for future calls to `navigate_1d` or
+    #: `navigate_grid`.
+    NONE = 0
+
+    #: Learn the menu structure to speed up future calls to `navigate_1d` or
+    #: `navigate_grid` during this test-run (that is, during the lifetime of
+    #: the Python process).
+    TEMPORARY = 1
+
+    #: Learn the menu structure and save it to a persistent cache on the
+    #: Stb-tester Node so that it can be used to speed up future calls to
+    #: `navigate_1d` or `navigate_grid` when running tests on the same
+    #: Stb-tester Node.
+    PERSISTENT = 2
+
+
+FrameObjectT = TypeVar("FrameObjectT", bound=FrameObject)
+
+
+class NavigationFailed(AssertionError):
+    """Raised by `navigate_1d` or `navigate_grid` if it couldn't find the
+    requested target.
+    """
+
+
+@dataclass(frozen=True, eq=False)
+class NavigationState:
+    """State of the device-under-test during navigation.
+
+    Passed to the ``recover`` parameter of `navigate_1d` and `navigate_grid`.
+    """
+
+    #: Latest frame of video from the device-under-test — that is, its current
+    #: state.
+    frame: Frame
+
+    #: The last Page Object seen when the page was still visible (before
+    #: navigation "fell off" the edge of the menu onto a different page).
+    last_page: FrameObject
+
+    #: The last key pressed during navigation, that caused it to "fall off" the
+    #: edge.
+    last_key: str
+
+
+def navigate_1d(
+        target: str,
+        page: FrameObjectT,
+        direction: Direction,
+        *,
+        mask: MaskTypes = Region.ALL,
+        timeout_secs: float = 120,
+        get_focus: Optional[Callable[[FrameObjectT], str]] = None,
+        eq: Callable[[str, str], bool] = ocr_eq,
+        press_and_wait: Optional[Callable[[str], Transition | TransitionStatus]] = None,
+        recover: Optional[str | Callable[[NavigationState], FrameObjectT]] = None,
+        retry_missed_keypresses: bool = False,
+        learning: Learning = Learning.PERSISTENT,
+        id: Optional[str] = None,  # pylint:disable=redefined-builtin
+        ) -> FrameObjectT:
+    """
+    Navigate through a 1-dimensional menu, searching for a target.
+
+    This doesn't press *KEY_OK* on the target; it only moves the focus there.
+
+    ``navigate_1d`` takes a ``page`` parameter, which is a Page Object that
+    describes the current state of the menu. The Page Object must have a
+    property called ``focus``. ``navigate_1d`` will search the menu until the
+    current focus (according to the Page Object) matches ``target``.
+
+    Typically this function is used from a Page Object's *navigate* method;
+    don't use this directly in your test scripts. For example::
+
+        class Menu(stbt.FrameObject):
+            @property
+            def is_visible(self):
+                ...  # implementation not shown
+
+            @property
+            def focus(self):
+                '''Read the currently focused menu item.'''
+                return stbt.ocr(frame=self._frame, region=...)
+
+            def navigate_to(self, target):
+                return stbt.navigate_1d(target, page=self,
+                                        direction=stbt.Direction.HORIZONTAL)
+
+    :param target: Navigate to this menu item. Typically a string.
+
+    :param page: An instance of a `stbt.FrameObject` sub-class that represents
+        the current state of the menu. We will use this object's ``focus``
+        property to determine which item is currently focused. After we have
+        pressed up, down, left, or right, we will call this object's
+        `refresh <stbt.FrameObject.refresh>` method to read the new state from
+        the screen.
+
+    :param direction: Direction to move in: Either `Direction.VERTICAL` (for
+        navigation with "KEY_DOWN" and "KEY_UP") or `Direction.HORIZONTAL` (for
+        navigation with "KEY_RIGHT" and "KEY_LEFT").
+
+    :param mask: A mask to use when calling `stbt.press_and_wait` to determine
+        when the current focus has finished moving. Usually this will be a
+        `Region` specifying where the menu is on the screen, but it can also be
+        a more complex :doc:`mask <masks>`.
+
+    :param timeout_secs: Raise `NavigationFailed` if we still haven't found the
+        target after this many seconds.
+
+    :param get_focus: By default, ``navigate_1d`` will use the ``page`` object's
+        ``focus`` property to determine which item is currently focused. You can
+        override this behaviour by providing a function that takes a single
+        argument (the ``page`` instance) and returns the focused menu item.
+
+    :param eq: A function that compares two strings (the current focus and the
+        target), and returns True if they match. The default is `stbt.ocr_eq`,
+        which ignores common OCR errors.
+
+    :param press_and_wait: ``navigate_1d`` will call `stbt.press_and_wait` with
+        the appropriate key ("KEY_DOWN", "KEY_UP", "KEY_LEFT", or "KEY_RIGHT")
+        to move the focus and wait until it has finished moving. If you need to
+        customise the parameters to `press_and_wait`, you can pass in your own
+        function here: It should accept a key name and call the real
+        `stbt.press_and_wait` with your extra parameters. For example::
+
+            def my_press_and_wait(key):
+                return stbt.press_and_wait(
+                    key, mask=stbt.Region(...), timeout_secs=10)
+            ...
+            return stbt.navigate_1d(
+                target, page=self, direction=stbt.Direction.HORIZONTAL,
+                press_and_wait=my_press_and_wait)
+
+        In many cases a `lambda` (anonymous function) is more convenient::
+
+            return stbt.navigate_1d(
+                target, page=self, direction=stbt.Direction.HORIZONTAL,
+                press_and_wait=lambda key: stbt.press_and_wait(
+                    key, mask=stbt.Region(...), timeout_secs=10))
+
+    :param recover: A function that will recover if navigation "falls off" the
+        edge of the menu onto a different page.
+
+        "Falling off" means that ``page.is_visible`` is False after navigating
+        up, down, left or right.
+
+        By default, ``navigate_1d`` will try to recover by pressing the
+        opposite key (for example, "KEY_LEFT" if it had fallen off after
+        pressing "KEY_RIGHT"), and if that fails, it will try "KEY_BACK". To use
+        a different recovery strategy, specify your own function here.
+
+        The function should take a single argument of type
+        `stbt.NavigationState`. It should do the appropriate actions to get back
+        onto the menu that we were navigating, and then return a new Page Object
+        that represents the new state of the menu. For example::
+
+            def my_recover(state: stbt.NavigationState):
+                assert stbt.press_and_wait("KEY_BACK")
+                return Menu()
+
+            stbt.navigate_1d(..., recover=my_recover)
+
+        If necessary, you can check the function's parameter to see the previous
+        Page Object (before falling off) and the last key that was pressed.
+
+        Your recovery function doesn't have to get back to the same menu item
+        that was focused before falling off; it just needs to get back onto the
+        menu. For example, if your Page Object has a staticmethod called
+        ``open`` that will open the menu from any state, you can use that to
+        recover, like this::
+
+            stbt.navigate_1d(..., recover=lambda _: Menu.open())
+
+        The ``recover`` parameter can also be a string, which is the name of
+        a key for `stbt.press`. ``navigate_1d`` will attempt to recover by
+        pressing that key once.
+
+    :param learning: We can remember the items we discovered during navigation
+        so that the next time you call ``navigate_1d`` we can use that
+        information to navigate faster. To disable this behaviour, pass
+        ``learning=stbt.Learning.NONE``. See `stbt.Learning`.
+
+        If the actual menu of the device-under-test doesn't match the learned
+        menu structure, this function will still work — it will fall back to the
+        searching behaviour.
+
+    :param id: A unique identifier for this menu. You only need to specify this
+        if ``learning`` is enabled and you are using the same FrameObject class
+        to recognize many different menus and sub-menus.
+
+    :returns: A new FrameObject instance of the same type as ``page``,
+        reflecting the device-under-test's new state after the navigation
+        completed.
+
+    :raises NavigationFailed: If we can't find the target.
+
+    Added in v34.
+    """
+    _raise_premium("navigate_1d")
+    return page.refresh()
+
+
+def navigate_grid(
+        target: str,
+        page: FrameObjectT,
+        *,
+        mask: MaskTypes = Region.ALL,
+        timeout_secs: float = 300,
+        get_focus: Optional[Callable[[FrameObjectT], str]] = None,
+        eq: Callable[[str, str], bool] = ocr_eq,
+        press_and_wait: Optional[Callable[[str], Transition | TransitionStatus]] = None,
+        recover: Optional[str | Callable[[NavigationState], FrameObjectT]] = None,
+        retry_missed_keypresses: bool = False,
+        learning: Learning = Learning.PERSISTENT,
+        id: Optional[str] = None,  # pylint:disable=redefined-builtin
+        ) -> FrameObjectT:
+    """
+    Navigate through a 2-dimensional grid, searching for a target.
+
+    This doesn't press *KEY_OK* on the target; it only moves the focus there.
+
+    See `navigate_1d` for documentation on all the parameters.
+
+    Added in v34.
+    """
+    _raise_premium("navigate_grid")
+    return page.refresh()
