@@ -61,10 +61,10 @@ class Counter:
     .. _instrumentation best practices: https://prometheus.io/docs/practices/instrumentation/#counter-vs-gauge-summary-vs-histogram
     .. _Metric names: https://prometheus.io/docs/practices/naming/
     """
-    def __init__(self, name, description):
+    def __init__(self, name: str, description: str):
         pass
 
-    def inc(self, value=1, labels=None):
+    def inc(self, value: float = 1, labels: "dict[str, str] | None" = None):
         """Increment the Counter by the given amount.
 
         :param int value: The amount to increase.
@@ -102,10 +102,10 @@ class Gauge:
 
     Added in v32.
     """
-    def __init__(self, name, description):
+    def __init__(self, name: str, description: str):
         pass
 
-    def set(self, value, labels=None):
+    def set(self, value: float, labels: "dict[str, str] | None" = None):
         """Set the Gauge to the given value.
 
         :param float value: The measurement to record.
@@ -147,10 +147,10 @@ class Histogram:
 
     Added in v32.
     """
-    def __init__(self, name, description, buckets):
+    def __init__(self, name: str, description: str, buckets: "list[float]"):
         pass
 
-    def log(self, value, labels=None):
+    def log(self, value: float, labels: "dict[str, str] | None" = None):
         """Store the given value into the Histogram.
 
         :param float value: The measurement to record.
